package ru.dbotthepony.kstarbound.defs.monster

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableMap
import com.google.gson.JsonElement
import com.google.gson.JsonNull
import com.google.gson.JsonObject
import ru.dbotthepony.kommons.io.readBinaryString
import ru.dbotthepony.kommons.io.writeBinaryString
import ru.dbotthepony.kommons.math.RGBAColor
import ru.dbotthepony.kommons.util.Either
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.client.render.RenderLayer
import ru.dbotthepony.kstarbound.defs.ActorMovementParameters
import ru.dbotthepony.kstarbound.defs.AssetPath
import ru.dbotthepony.kstarbound.defs.ClientEntityMode
import ru.dbotthepony.kstarbound.defs.ColorReplacements
import ru.dbotthepony.kstarbound.defs.JsonFunction
import ru.dbotthepony.kstarbound.defs.TeamType
import ru.dbotthepony.kstarbound.defs.actor.StatusControllerConfig
import ru.dbotthepony.kstarbound.defs.animation.AnimationDefinition
import ru.dbotthepony.kstarbound.defs.item.TreasurePoolDefinition
import ru.dbotthepony.kstarbound.fromJsonFast
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.json.readJsonElement
import ru.dbotthepony.kstarbound.json.writeJsonElement
import ru.dbotthepony.kstarbound.math.AABB
import ru.dbotthepony.kstarbound.math.vector.Vector2d
import ru.dbotthepony.kstarbound.util.random.random
import ru.dbotthepony.kstarbound.util.random.staticRandomInt
import ru.dbotthepony.kstarbound.world.PIXELS_IN_STARBOUND_UNIT
import ru.dbotthepony.kstarbound.world.physics.Poly
import java.io.DataInputStream
import java.io.DataOutputStream
import java.util.random.RandomGenerator

@JsonFactory
class MonsterVariant(
	val type: String,
	val shortDescription: String? = null,
	val description: String? = null,
	val seed: Long = 0L,
	val uniqueParameters: JsonObject = JsonObject(),
	val animationConfig: AnimationDefinition,
	val reversed: Boolean = false,
	val animatorPartTags: ImmutableMap<String, String>,
	val parameters: JsonObject,

	@Deprecated("Raw property", replaceWith = ReplaceWith("this.actualDropPools"))
	val dropPools: ImmutableList<Either<ImmutableMap<String, Registry.Ref<TreasurePoolDefinition>>, Registry.Ref<TreasurePoolDefinition>>> = ImmutableList.of(),
) {
	@JsonFactory
	data class CommonParameters(
		val shortDescription: String? = null,
		val description: String? = null,
		val scripts: ImmutableList<AssetPath>,
		val animationScripts: ImmutableList<AssetPath> = ImmutableList.of(),
		val animationCustom: JsonElement = JsonNull.INSTANCE,
		val initialScriptDelta: Double = 5.0,
		val metaBoundBox: AABB,
		val scale: Double,
		val renderLayer: RenderLayer.Point = RenderLayer.Point(RenderLayer.Monster),
		val movementSettings: ActorMovementParameters = ActorMovementParameters.EMPTY,

		val dropPools: ImmutableList<Either<ImmutableMap<String, Registry.Ref<TreasurePoolDefinition>>, Registry.Ref<TreasurePoolDefinition>>>? = null,

		val walkMultiplier: Double = 1.0,
		val runMultiplier: Double = 1.0,
		val jumpMultiplier: Double = 1.0,
		val weightMultiplier: Double = 1.0,
		val healthMultiplier: Double = 1.0,
		val touchDamageMultiplier: Double = 1.0,
		val touchDamage: JsonElement = JsonNull.INSTANCE,
		val animationDamageParts: JsonObject = JsonObject(),
		val statusSettings: StatusControllerConfig? = null,
		val knockoutTime: Double = 1.0,
		val knockoutEffect: String? = null,
		val knockoutAnimationStates: ImmutableMap<String, String> = ImmutableMap.of(),

		// in pixels
		val mouthOffset: Vector2d,

		// in pixels
		val feetOffset: Vector2d,

		val powerLevelFunction: Registry.Ref<JsonFunction> = Registries.jsonFunctions.ref("monsterLevelPowerMultiplier"),
		val healthLevelFunction: Registry.Ref<JsonFunction> = Registries.jsonFunctions.ref("monsterLevelHealthMultiplier"),

		val clientEntityMode: ClientEntityMode = ClientEntityMode.CLIENT_SLAVE_ONLY,
		val persistent: Boolean = false /* WHAT. */,
		val damageTeamType: TeamType = TeamType.ENEMY,
		val damageTeam: Int = 2,

		val selfDamagePoly: Poly = movementSettings.standingPoly?.map({ it }, { it.firstOrNull() }) ?: Poly.EMPTY,
		val portraitIcon: String? = null,

		val damageReceivedAggressiveDuration: Double = 1.0,
		val onDamagedOthersAggressiveDuration: Double = 5.0,
		val onFireAggressiveDuration: Double = 5.0,

		val nametagColor: RGBAColor = RGBAColor.WHITE,
		val colorSwap: ColorReplacements? = null,
	) {
		val mouthOffsetTiles = mouthOffset / PIXELS_IN_STARBOUND_UNIT
		val feetOffsetTiles = feetOffset / PIXELS_IN_STARBOUND_UNIT
	}

	val entry = Registries.monsterTypes.getOrThrow(type)
	val commonParameters = Starbound.gson.fromJsonFast(parameters, CommonParameters::class.java)

	val animatorZoom: Double
		get() = commonParameters.scale

	val actualDropPools: ImmutableList<Either<ImmutableMap<String, Registry.Ref<TreasurePoolDefinition>>, Registry.Ref<TreasurePoolDefinition>>>
		get() = commonParameters.dropPools ?: dropPools

	val chosenDropPool: Either<ImmutableMap<String, Registry.Ref<TreasurePoolDefinition>>, Registry.Ref<TreasurePoolDefinition>>? = if (actualDropPools.isEmpty()) null else actualDropPools[staticRandomInt(0, actualDropPools.size, seed, "MonsterDropPool")]

	fun write(stream: DataOutputStream, isLegacy: Boolean) {
		if (isLegacy) {
			stream.writeBinaryString(type)
			stream.writeLong(seed)
			stream.writeJsonElement(uniqueParameters)
		} else {
			TODO("Native")
		}
	}

	companion object {
		fun read(stream: DataInputStream, isLegacy: Boolean): MonsterVariant {
			if (isLegacy) {
				val name = stream.readBinaryString()
				val seed = stream.readLong()
				val uniqueParameters = stream.readJsonElement() as JsonObject

				val base = Registries.monsterTypes.getOrThrow(name)
				return base.value.create(seed, uniqueParameters)
			} else {
				TODO("Native")
			}
		}
	}
}
