package ru.dbotthepony.kstarbound.world.entities.behavior

import com.google.common.collect.ImmutableList
import com.google.gson.JsonArray
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import ru.dbotthepony.kommons.gson.contains
import ru.dbotthepony.kommons.gson.get
import ru.dbotthepony.kommons.gson.stream
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.actor.behavior.CompositeNodeType
import ru.dbotthepony.kstarbound.defs.actor.behavior.NodeParameterValue
import ru.dbotthepony.kstarbound.fromJsonFast
import ru.dbotthepony.kstarbound.lua.userdata.BehaviorState
import ru.dbotthepony.kstarbound.util.valueOf

abstract class AbstractBehaviorNode {
	enum class Status(val asBoolean: Boolean?) {
		INVALID(null),
		SUCCESS(true),
		FAILURE(false),
		RUNNING(null)
	}

	abstract fun run(delta: Double, state: BehaviorState): Status
	abstract fun reset()

	fun runAndReset(delta: Double, state: BehaviorState): Status {
		val status = run(delta, state)

		if (status != Status.RUNNING)
			reset()

		return status
	}

	companion object {
		fun replaceBehaviorTag(parameter: NodeParameterValue, treeParameters: Map<String, NodeParameterValue>): NodeParameterValue {
			var str: String? = null

			if (parameter.key != null)
				str = parameter.key

			// original engine does this, and i don't know why this make any sense
			else if (parameter.value is JsonPrimitive && parameter.value.isString)
				str = parameter.value.asString

			if (str != null) {
				if (str.first() == '<' && str.last() == '>') {
					val treeKey = str.substring(1, str.length - 2)
					val param = treeParameters[treeKey]

					if (param != null) {
						return param
					} else {
						throw NoSuchElementException("No parameter specified for tag '$str'")
					}
				}
			}

			return parameter
		}

		fun replaceOutputBehaviorTag(output: String?, treeParameters: Map<String, NodeParameterValue>): String? {
			if (output == null) {
				return null
			} else if (output.first() == '<' && output.last() == '>') {
				val replacement = treeParameters[output.substring(1, output.length - 2)] ?: throw NoSuchElementException("No parameter specified for tag '$output'")

				if (replacement.key != null)
					return replacement.key
				else if (replacement.value is JsonPrimitive && replacement.value.isString)
					return replacement.value.asString
				else
					return null
			} else {
				return output
			}
		}

		fun create(data: JsonObject, treeParameters: Map<String, NodeParameterValue>, tree: BehaviorTree): AbstractBehaviorNode {
			val type = BehaviorNodeType.entries.valueOf(data["type"].asString)
			val name = data["name"].asString
			val parameterConfig = data.get("parameters") { JsonObject() }

			if (type == BehaviorNodeType.MODULE) {
				// merge in module parameters to a copy of the treeParameters to propagate
				// tree parameters into the sub-tree, but allow modules to override
				val moduleParameters = LinkedHashMap(treeParameters)

				for ((k, v) in parameterConfig.entrySet()) {
					moduleParameters[k] = replaceBehaviorTag(Starbound.gson.fromJsonFast(v, NodeParameterValue::class.java), treeParameters)
				}

				val module = BehaviorTree(tree.blackboard, Registries.behavior.getOrThrow(name).value, moduleParameters)
				tree.scripts.addAll(module.scripts)
				tree.functions.addAll(module.functions)
				return tree.root
			}

			val parameters = LinkedHashMap(Registries.behaviorNodes.getOrThrow(name).value.properties)

			for ((k, v) in parameters.entries) {
				if (k in parameterConfig) {
					parameters[k] = v.copy(value = replaceBehaviorTag(Starbound.gson.fromJsonFast(parameterConfig[k], NodeParameterValue::class.java), treeParameters))
				} else {
					val replaced = replaceBehaviorTag(v.value, treeParameters)

					if (replaced != v.value) {
						parameters[k] = v.copy(value = replaced)
					}
				}
			}

			when (type) {
				BehaviorNodeType.ACTION -> {
					tree.functions.add(name)

					val outputConfig = data.get("output") { JsonObject() }
					val output = LinkedHashMap(Registries.behaviorNodes.getOrThrow(name).value.output)

					for ((k, v) in output.entries) {
						val replaced = replaceOutputBehaviorTag(outputConfig[k]?.asString ?: v.key, treeParameters)

						if (replaced != v.key) {
							output[k] = v.copy(key = replaced)
						}
					}

					return ActionNode(name, parameters, output)
				}

				BehaviorNodeType.DECORATOR -> {
					tree.functions.add(name)
					val sacrifice = create(data["child"] as JsonObject, treeParameters, tree)
					return DecoratorNode(name, parameters, sacrifice)
				}

				BehaviorNodeType.COMPOSITE -> {
					val children = data.get("children") { JsonArray() }
						.stream()
						.map { create(it as JsonObject, treeParameters, tree) }
						.collect(ImmutableList.toImmutableList())

					return CompositeNodeType.entries.valueOf(name).factory(parameters, children)
				}

				BehaviorNodeType.MODULE -> throw RuntimeException()
			}
		}
	}
}
